#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/boot.h>                                                   //boot_lock_fuse_bits_get
#include <arduino.h>                                                    //Need millis etc.
#include "myStepperDriver.h"

                                                                        //IRQ Driver:
uint32_t      motor_millis_endtime_irq;                                 //Time (millis) the motor movement has to stop. 0 = no end time.
uint32_t      motor_step_count_limit_irq;                               //Step count limit when the motor movement has to stop. 0 = no step limit.
uint32_t      motor_time_irq;                                           //Time (millis) variable speed has to be updated.
uint32_t      motor_timer_top32virt_irq;                                //Virtual 32bit TOP register for Timer1.
uint32_t      motor_timer_modify_irq;                                   //Virtual 32bit modification value for Timer1.
uint16_t      motor_update_time_irq;                                    //Interval (ms) between speed updates.
void          (*motorCBreached) ();                                     //Callback function.
void          (*motorCBfault) ();                                       //Callback function.
uint16_t      motor_modCounts_irq;                                      //While !=0, allows motor speed modifications.
uint8_t       motor_ctrl_irq;                                           //Preserves the control byte from the \ref motorSetSpeed function.
uint8_t       motor_stat_irq;                                           //Motor status output. Can be read using \ref motorReadStatus function.
uint8_t       motor_pattern_irq;                                        //Only the 2 lowest bits are used: step motor pattern. Other bits are kept zero.
uint8_t       skip_count_irq;                                           //Computed value how many Timer1 IRQs can be skipped before executing \ref motorTimerUpdateCheck function.
                                                                        //Only valid when MOTOR_CTRL_SPEED_VARIABLE is included in \ref motorSetSpeed function call.

static void   timerUpdate (uint32_t timTop32);                          //Updates HW Timer1 from virtual 32bit timer TOP value.
static void   timerSkipCount (uint32_t virtual32bitTimer);              //Computes how many Timer1 IRQs fit in 50ms interval to define skip_count_irq.


#ifdef DEBUG_BAUDRATE                                                   //This is some debugging stuff, not to be used in the final application as it uses the serial port.
  uint8_t     serialInitialized;
  /********************************************************************************************************
   */
  void
  serDbgPrint (const char* formatString,                                //Output string. May contain a single %i for printing a byte (function may become ...).
                           ...)                                         //Optional data to be printed.
  {
    if (formatString == NULL) return;
    va_list args;
    char text[256];
    va_start (args, formatString);
    vsnprintf (text, sizeof(text), formatString, args);
    va_end (args);
    Serial.println (text);
    delay (10);
  }





  void
  debugDump (bool only_if_speed_variable)
  {
    if (   (only_if_speed_variable != 0)
        && ((motor_ctrl_irq & MOTOR_CTRL_IRQ_TIMING_VARIABLE) == 0)) return;
    DEBUGPRINT (("motor_ctrl_irq: 0x%02X", motor_ctrl_irq));
    DEBUGPRINT (("motor_timer_top32virt_irq: %lu", motor_timer_top32virt_irq));
    DEBUGPRINT (("motor_timer_modify_irq: %lu", motor_timer_modify_irq));
    DEBUGPRINT (("motor_step_count_limit_irq: %lu", motor_step_count_limit_irq));
    DEBUGPRINT (("motor_time_irq: %lu", motor_time_irq));
    DEBUGPRINT (("motor_stat_irq: 0x%02X", motor_stat_irq));
    DEBUGPRINT (("motor_update_time_irq: %u", motor_update_time_irq));
    DEBUGPRINT (("motor_millis_endtime_irq: %lu", motor_millis_endtime_irq));
    DEBUGPRINT (("motor_modCounts_irq: %u", motor_modCounts_irq));
    DEBUGPRINT (("motor_pattern_irq: %u", motor_pattern_irq));
  }
#endif





#ifdef MOTOR_IOPORT_A
  #define MOTOR_DDRPORT  DDRA
  #define MOTOR_IOPORT   PORTA
#endif
#ifdef MOTOR_IOPORT_B
  #define MOTOR_DDRPORT  DDRB
  #define MOTOR_IOPORT   PORTB
#endif
#ifdef MOTOR_IOPORT_C
  #define MOTOR_DDRPORT  DDRC
  #define MOTOR_IOPORT   PORTC
#endif
#ifdef MOTOR_IOPORT_D
  #define MOTOR_DDRPORT  DDRD
  #define MOTOR_IOPORT   PORTD
#endif
#ifdef MOTOR_IOPORT_E
  #define MOTOR_DDRPORT  DDRE
  #define MOTOR_IOPORT   PORTE
#endif
#ifdef MOTOR_IOPORT_F
  #define MOTOR_DDRPORT  DDRF
  #define MOTOR_IOPORT   PORTF
#endif
#ifdef MOTOR_IOPORT_G
  #define MOTOR_DDRPORT  DDRG
  #define MOTOR_IOPORT   PORTG
#endif
#ifdef MOTOR_IOPORT_H
  #define MOTOR_DDRPORT  DDRH
  #define MOTOR_IOPORT   PORTH
#endif
#ifdef MOTOR_IOPORT_J
  #define MOTOR_DDRPORT  DDRJ
  #define MOTOR_IOPORT   PORTJ
#endif
#ifdef MOTOR_IOPORT_K
  #define MOTOR_DDRPORT  DDRK
  #define MOTOR_IOPORT   PORTK
#endif
#ifdef MOTOR_IOPORT_L
  #define MOTOR_DDRPORT  DDRL
  #define MOTOR_IOPORT   PORTL
#endif





/********************************************************************************************************
 *    \ingroup LowLevel
 *    Interrupt service routine for TIMER, used in FAST PWM mode to apply a single motor step.
 *    Note: if the endpoint has been reached (either by timed motor operation or step count limit), Timer interrupts are disabled, but the driver
 *    remains activated. To conserve power, it's up to the user to use \ref motorReadStatus and determine when movement has stopped and turn off the
 *    driver when the conditions permit.
 */
#ifdef MOTOR_TIMER1
  ISR (TIMER1_OVF_vect)                                                 //Timer/Counter1 Overflow.
#endif
#ifdef MOTOR_TIMER3
  ISR (TIMER3_OVF_vect)                                                 //Timer/Counter3 Overflow.
#endif
#ifdef MOTOR_TIMER4
  ISR (TIMER4_OVF_vect)                                                 //Timer/Counter4 Overflow.
#endif
#ifdef MOTOR_TIMER5
  ISR (TIMER5_OVF_vect)                                                 //Timer/Counter5 Overflow.
#endif
{
  if (motor_ctrl_irq & MOTOR_CTRL_DIRECTION) --motor_pattern_irq;       //Reverse direction.
  else ++motor_pattern_irq;
  motor_pattern_irq &= 0x03;                                            //Only 4 patterns are possible.
  #ifdef ARDUINO_MOTOR_SHIELD
    unsigned char p = PORTB;
    #ifdef _ARDUINO_UNO_                     //76543210
      const unsigned char MOTOR_PATTERN[] = { B00000000,
                                              B00100000,
                                              B00110000,
                                              B00010000 };
      p &= 0xCF;                                                        //Clear pins PB4, PB5 to both zero.
    #endif
    #ifdef _ARDUINO_MEGA_
      const unsigned char MOTOR_PATTERN[] = { B00000000,
                                              B10000000,
                                              B11000000,
                                              B01000000 };
      p &= 0x3F;                                                         //Clear pins PB7, PB6 to both zero.
    #endif
    p |= MOTOR_PATTERN[motor_pattern_irq];                              //Define the new motor pattern in the PBx pins.
    PORTB = p;                                                          //Update the I/O port with the new output state.
  #else
    unsigned char p = MOTOR_IOPORT;
    p &= ~MOTOR_OUTPUTS;
    const unsigned char MOTOR_PATTERN[] = { MOTOR_PATTERN0,
                                            MOTOR_PATTERN1,
                                            MOTOR_PATTERN2,
                                            MOTOR_PATTERN3 };
    p |= MOTOR_PATTERN[motor_pattern_irq];                              //Define the new motor pattern in the port pins.
    MOTOR_IOPORT = p;
  #endif
  if ((motor_ctrl_irq & MOTOR_CTRL_IRQ_TIMING_VARIABLE) == MOTOR_CTRL_IRQ_TIMING_VARIABLE)//Timing IRQ + Variable.
  {
    if (skip_count_irq != 0) --skip_count_irq;                          //If timer1 has a low divider value, skipcount is set vs. the divider to avoid too many
    else motorTimerUpdateCheck ();                                      //timer TOP update (if needed).
  }
  if (motor_step_count_limit_irq != 0)                                  //Only executed if a motor stepcount limit is defined
  {
    --motor_step_count_limit_irq;                                       //Reduce the amount of motor steps to be applied.
    if (motor_step_count_limit_irq == 0) motor_stat_irq |= MOTOR_STAT_REACHED;//Flag end of steps has been reached.
  }
  if (   (motor_millis_endtime_irq != 0)                                //If user used stepcount limit as time limit, it was stored in here.
      && (millis () > motor_millis_endtime_irq)) motor_stat_irq |= MOTOR_STAT_REACHED;//Flag motor revolutions stop by endtime settling.
  if (motor_stat_irq & MOTOR_STAT_REACHED)
  {
    motor_stat_irq &= ~MOTOR_STAT_SPINNING;                             //Motor operation stopped.
    #ifdef MOTOR_TIMER1
      TIMSK1 &= ~(1 << TOIE1);                                            //Disable Timer1 Overflow Interrupt Enable: motor step count has been reached.
    #endif
    #ifdef MOTOR_TIMER3
      TIMSK3 &= ~(1 << TOIE3);                                          //Disable Timer3 Overflow Interrupt Enable: motor step count has been reached.
    #endif
    #ifdef MOTOR_TIMER4
      TIMSK4 &= ~(1 << TOIE4);                                          //Disable Timer4 Overflow Interrupt Enable: motor step count has been reached.
    #endif
    #ifdef MOTOR_TIMER5
      TIMSK5 &= ~(1 << TOIE5);                                          //Disable Timer5 Overflow Interrupt Enable: motor step count has been reached.
    #endif
  }                                                                     //Note: this will NOT turn off the drivers.
}





/********************************************************************************************************
 *    \ingroup API
 *    The user can define "void motorFault ()" function.
 *    This function gets called when a motor fault / controlling error is encountered and the timer TOP value is updated
 *    by frequent calls of motorTimerUpdateCheck.
 *    @param[in] userFunction         Name of a function declared 'void myFunction ()' that handles the error event.
 */
void
motorCallBackFault (void (*userFunction) ())
{
  motorCBfault = userFunction;
}





/********************************************************************************************************
 *    \ingroup API
 *    This function can be defined by the user. It's called when the motor reaches it's end position. This allows the
 *    user to initiate motor rotation and no longer care about it. Once the stepcount limit has been reached, the
 *    motor is stopped and the user defined function is called to make the software aware about the end position.
 *    @param[in] userFunction         Name of a function declared 'void myFunction ()' that handles the position reached event.
 */
void
motorCallBackPositionReached (void (*userFunction) ())
{
  motorCBreached = userFunction;
}





/********************************************************************************************************
 *    \ingroup API
 *    If the event functions are not used, the user application can also poll the motor status.
 *    \retval                     Returns any combination of MOTOR_STAT_* bits.
 */
uint8_t
motorReadStatus ()
{
  return motor_stat_irq;
}






/********************************************************************************************************
 *    \ingroup API
 *    Must be called once before starting the motor operation.
 */
void
motorShieldInit ()
{
  #ifdef ARDUINO_MOTOR_SHIELD
    pinMode (     MOTOR_ENABLE_A, OUTPUT);
    pinMode (      MOTOR_BREAK_B, OUTPUT);
    pinMode (      MOTOR_BREAK_A, OUTPUT);
    pinMode (     MOTOR_ENABLE_B, OUTPUT);
    pinMode (        MOTOR_DIR_A, OUTPUT);
    pinMode (        MOTOR_DIR_B, OUTPUT);
    digitalWrite (MOTOR_ENABLE_A,    LOW);
    digitalWrite ( MOTOR_BREAK_B,    LOW);
    digitalWrite ( MOTOR_BREAK_A,    LOW);
    digitalWrite (MOTOR_ENABLE_B,    LOW);
    digitalWrite (   MOTOR_DIR_A,    LOW);
    digitalWrite (   MOTOR_DIR_B,    LOW);
  #else
    MOTOR_DDRPORT = MOTOR_OUTPUTS;
    MOTOR_IOPORT = ~MOTOR_OUTPUTS;
  #endif
}







/********************************************************************************************************
 *    \ingroup API
 *    Compute a 32bit virtual timer TOP value to use with Fast PWM mode. A stepper motor spins by applying step pulses.
 *    These pulses must be generated at a fixed interval. This driver uses the 16-bit hardware TIMER to divide the main
 *    clock (obtained by calling motorTimerGetIOclocksPerSecond) by the amount of step one wants to send to the motor.
 *    TIMER than takes care to redistribute the steps evenly over 1 second.
 *    The minimal steps/s based on this call is 1. Should one decide to let the motor run even slower, one may consider
 *    to use motorTimerCalcPerMinute.
 *    @param[in]  motorPulses         Amount of motor steps per second. Zero is not allowed.
 *    @retval                         Calculated virtual timer TOP as a 32-bit IO-clock count.
 */
uint32_t
motorTimerCalcPerSecond (uint16_t motorPulses)
{
  uint32_t clk = motorTimerGetIOclocksPerSecond ();
  if (motorPulses != 0) clk /= (uint32_t) (motorPulses);                //Calculate the timer TOP value as a virtual 32-bit value.
  return clk;                                                           //Timer virtual TOP register of 32 bits.
}





/********************************************************************************************************
 *    \ingroup API
 *    Returns the IO clock count per second.
 *    The AVR core has an 8MHz RC or 128kHz system oscillator. These can be selected via programmable 'Fuses'.
 *    The external oscillator in Arduino is 16MHz and believed to be the major oscillator source. The function
 *    first determines the clock source used. Then, it checks the CLKPR (Clock Prescaler) register in case the
 *    user has reduced the speed to preserve energy. That results in a final amount of I/O clocks available to the application.
 *    \retval                     I/O clocks per second.
 */
uint32_t
motorTimerGetIOclocksPerSecond ()
{
  uint32_t ioClk = F_CPU;                                               //Expected 16MHz.
  switch (boot_lock_fuse_bits_get (0) & 0x0F)                           //Read the AVR clock programming fuses. Unless internal clock is active,
  {                                                                     //F_CPU is most likely the correct CPU clock frequency.
  case   3:     ioClk =  128000UL; break;                               //Running at 128kHz system clock.
  case   2:     ioClk = 8000000UL; break;                               //8MHz RC.
  default:      break;                                                  //F_CPU expected to be 16000000;
  }
  char clkpr = CLKPR;                                                   //The clock prescaler register allows to divide the real clock to reduce power.
  if (clkpr & 0x80)                                                     //Clock prescaler enabled?
  {
    clkpr &= 0x0F;                                                      //Only the 4 lowest bits are valid.
    if (clkpr < 0x09) while (clkpr)                                     //Clock divider 0 ... 8 are valid.
    {
      ioClk >>= 1;                                                      //Divide IOclock by 2.
      --clkpr;
    }
  }
  return ioClk;
}





/********************************************************************************************************
 *    \ingroup API
 *    The user should call this timer update function about every 1-20ms if variable motor speed is active.
 *    If this cannot be guaranteed, the user can set the MOTOR_CTRL_IRQ_TIMING bit in the control byte when calling
 *    \ref motorSetSpeed function. The interrupt service routine will than call this update function on its own,
 *    but timing will not be perfect (to reduce interrupt overhead).
 */
void
motorTimerUpdateCheck ()
{
  if (   ((motor_ctrl_irq & MOTOR_CTRL_IRQ_TIMING) == 0)                //Not under interrupt control (don't want potential blocking user code).
      && ((motor_stat_irq & (  MOTOR_STAT_REACHED
                             | MOTOR_STAT_REACH_CALLED)) == MOTOR_STAT_REACHED)//Motor Position has been reached but not called.
      && (motorCBreached != NULL))
  {
    motor_stat_irq |= MOTOR_STAT_REACH_CALLED;                          //Remind the function has been called once.
    (*motorCBreached) ();                                               //Call user attention 'motor has been stopped at target position'.
  }
  if ((motor_ctrl_irq & MOTOR_CTRL_SPEED_VARIABLE) == 0) return;        //Static motor speed (no need to check for timer variations).
  if (motor_timer_modify_irq == 0UL) return;                            //Timer Top modification == 0 -> Speed is static.
  uint32_t now = millis ();
  uint32_t elapsed = now - motor_time_irq;
  if ((uint16_t) (elapsed) < motor_update_time_irq) return;
  motor_time_irq += (uint32_t) (motor_update_time_irq);                 //Remind the Timer TOP update occurred relative to previous limit.
  uint8_t valid = 0;
  if ((motor_ctrl_irq & MOTOR_CTRL_FASTER) != 0)
  {                                                                     //Faster rotation: sub a value from the virtual timer TOP register.
    now = motor_timer_top32virt_irq - motor_timer_modify_irq;
    if (now < motor_timer_top32virt_irq) valid = 1;                     //Prevent sign inversion effects.
    if ((motor_ctrl_irq & MOTOR_CTRL_VARIABLE_PWR) != 0) motor_timer_modify_irq >>= 1;
  }
  else                                                                  //Slower rotation: add a value from the virtual timer TOP register.
  {                                                                     //ADD
    now = motor_timer_top32virt_irq + motor_timer_modify_irq;
    if (now > motor_timer_top32virt_irq) valid = 2;                     //Prevent sign inversion effets.
    if ((motor_ctrl_irq & MOTOR_CTRL_VARIABLE_PWR) != 0) motor_timer_modify_irq <<= 1;
  }
  if (valid == 0)                                                       //Could not update the virtual timer TOP.
  {
    motor_stat_irq |= MOTOR_STAT_FAULT;                                 //Remind the error; but rotation interrupt unchanged.
    if (   ((motor_ctrl_irq & ( MOTOR_CTRL_IRQ_TIMING
                              | MOTOR_STAT_FAULT_CALLED)) == 0)         //Not allowed when IRQ routine does timing to prevent IRQ blocking everything.
        && (motorCBfault != NULL))
    {
      motor_stat_irq |= MOTOR_STAT_FAULT_CALLED;                        //Remind the function has been called once.
      (*motorCBfault) ();                                               //Call user attention if defined.
    }
  }
  else                                                                  //"now" is the new 32bit virtual TOP register.
  {
    if (now < 640UL) now = 640UL;                                       //Minimal timer TOP value.
    motor_timer_top32virt_irq = now;                                    //New virtual 32bit Timer TOP register.
    timerUpdate (now);
    timerSkipCount (now);
    if (motor_timer_modify_irq == 0) motor_modCounts_irq = 1;
    if (motor_modCounts_irq != 0)                                       //Modificiation restriction counter active?
    {
      --motor_modCounts_irq;
      motor_stat_irq |= MOTOR_STAT_SPIN_FIXED;
      if (motor_modCounts_irq == 0) motor_ctrl_irq &= ~MOTOR_CTRL_IRQ_TIMING_VARIABLE;//From now on fixed speed.
    }
  }
}





/********************************************************************************************************
 *    \ingroup LowLevel
 *    If the virtual timer has higher interrupt rate (less than 50ms), we do not want to check the virtual timer update every time if the timer
 *    update is under local control (during timer IRQ processing). This to reduce timer overhead.
 *    In such case, we compute the amount of Timer updates that occurs per second. This is divided by 20 (resulting from the 50ms minimal timer update interval).
 *    That's the amount of IRQs skipped before timer update interval is allowed to execute.
 *    @param[in] virtual32bitTimer    Virtual 32bit timer to STEP the motor. This is the divider for the IO Clock.
 */
static void
timerSkipCount (uint32_t virtual32bitTimer)
{
  virtual32bitTimer = motorTimerCalcPerSecond (1) / virtual32bitTimer;  //Calculate IRQs per second from timer 1.
  skip_count_irq = 0;                                                   //Skip nothing.
  if (virtual32bitTimer > 20UL)                                         //Timer update minimal is 50ms -> timerUpd_ms limit in motorSetSpeed function.
  {
    virtual32bitTimer /= 50UL;
    if (virtual32bitTimer > 0xFFUL) virtual32bitTimer = 0xFFUL;         //The skip counter is 8-bit, so keep it limit to that maximal value.
    skip_count_irq = (uint8_t) (virtual32bitTimer);
  }
}






/********************************************************************************************************
 *    \ingroup LowLevel
 *    The virtual 32bit top value is recalculated to the nearest 16bit timer parameters using available divider.
 *    Since the timer hardware is 16bit, basically the real timer TOP value is 16 bit shifted vs. the divider option required
 *    to have best timer resolution vs. provided virtual 32bit timer TOP value.
 *    @param[in] timTop32             A virtual 32bit IO clock divider required to obtain the amount of wanted motor steps per second.
 *                                    Minimal value: 640 allowed. This is not tested. From HW point of view, lower is allowed but IRQ service
 *                                    routine will not be able to handle the IRQ rate using such low value.
 *                                    Maximal value: (1024 * 65536) - 1.
 *
 *    Example: the user calls this function with parameter 4159231 (0x3F76FF). The I/O clock is 16MHz.
 *    One would expect 16M / 4159231 = 3.846865 steps/s.
 *    Selected Prescaler    32bit Timer     Hardware 16bit Timer
 *           /1             4159231         does not fit
 *           /8              519903         does not fit
 *          /64               64987         0xFDD8
 *    The real speed will be 3.846923 steps/s.
 *    So the real divider is not 4159231, but 64 * 64987 = 4159168. This deviation results from the AVR core Timer 1: a 16bit hardware counter.
 *    Not all possible 32-bit value can be achieved since the prescaler limits accuracy for count values exceeding 16bit. However, the deviation is
 *    still less than 0.01% from the value provided by the user. Timer values below 0xFFFF will have zero deviation, since all fit in the 16bit Timer.
 *
 */
static void
timerUpdate (uint32_t timTop32)                                         //Virtual 32-bit timer data TOP value.
{
  unsigned char cs = (1 << CS10);                                       //Divider 1.
  if ((timTop32 & 0xFFFF0000UL) != 0)                                   //Doesn't fit in 16-bit.
  {
    timTop32 >>= 3;
    cs = (1 << CS11);                                                   //Divide by 8.
  }
  if ((timTop32 & 0xFFFF0000UL) != 0)                                   //Doesn't fit in 16-bit.
  {
    timTop32 >>= 3;
    cs = (1 << CS11) | (1 << CS10);                                     //Divide by 64.
  }
  if ((timTop32 & 0xFFFF0000UL) != 0)                                   //Doesn't fit in 16-bit.
  {
    timTop32 >>= 2;
    cs = (1 << CS12);                                                   //Divide by 256.
  }
  if ((timTop32 & 0xFFFF0000UL) != 0)                                   //Doesn't fit in 16-bit.
  {
    timTop32 >>= 2;
    cs = (1 << CS12) | (1 << CS10);                                     //Divide by 1024.
  }
  if ((timTop32 & 0xFFFF0000UL) != 0)
  {
    motor_stat_irq |= MOTOR_STAT_TIMER_DIV_ERROR;
    if (   ((motor_stat_irq & MOTOR_STAT_TIMER_DIV_CALLED) == 0)        //Fault Function not yet called.
        && ((motor_ctrl_irq & MOTOR_CTRL_IRQ_TIMING) == 0)             //Timing not under IRQ control.
        && (motorCBfault != NULL))                                      //Fault function defined by user.
    {
      motor_stat_irq |= MOTOR_STAT_TIMER_DIV_CALLED;                    //Remind function was called.
      (*motorCBfault) ();                                               //Call user attention if defined.
    }
  }
//if ((timTop32 == 0) timTop32 = 0xFFFF;                                //Make sure the timer register is never zero (increases IRQ overhead if called in IRQ routine).
  unsigned char lo = (unsigned char) (timTop32);                        //Timer TOP value lowest byte.
  timTop32 >>= 8;
  unsigned char hi = (unsigned char) (timTop32);                        //Timer1 TOP value highest byte.
  #ifdef MOTOR_TIMER1
    OCR1AH = hi;
    OCR1AL = lo;                                                        //Hardware update (high byte written first).
    TCCR1B = (1 << WGM13) | (1 << WGM12) | cs;                          //WGM must be %1111 + set/update clock divider.
  #endif
  #ifdef MOTOR_TIMER3
    OCR3AH = hi;
    OCR3AL = lo;                                                        //Hardware update (high byte written first).
    TCCR3B = (1 << WGM33) | (1 << WGM32) | cs;                          //WGM must be %1111 + set/update clock divider.
  #endif
  #ifdef MOTOR_TIMER4
    OCR4AH = hi;
    OCR4AL = lo;                                                        //Hardware update (high byte written first).
    TCCR4B = (1 << WGM43) | (1 << WGM42) | cs;                          //WGM must be %1111 + set/update clock divider.
  #endif
  #ifdef MOTOR_TIMER5
    OCR5AH = hi;
    OCR5AL = lo;                                                        //Hardware update (high byte written first).
    TCCR5B = (1 << WGM53) | (1 << WGM52) | cs;                          //WGM must be %1111 + set/update clock divider.
  #endif
}





/********************************************************************************************************
 *    \ingroup API
 *    @param[in]  control             MOTOR_CTRL* parameter(s).
 *    @param[in]  timerTop32virtual   A virtual 32-bit timer interval. Every time the timer reaches this value, 1 motor step is performed.
 *                                    The effective maximal motor speed must be recomputed vs. the actual I/O clock frequency (assumed this is 16MHz on these Arduino's).
 *                                    Allowed value: 0: stops the motor spin.
 *                                    Minimal value: 640 (this is the minimal amount of clocks allocated to decently process the interrupt service routine & other tasks).
 *                                    Maximal value: (1024 * 65536) - 1. Reason is the timer prescaler and 16-bit timer range.
 *                                    If the virtual 32bit TOP exceeds the maximal value, the motor is stopped.
 *                                    The function will search for the best fit Timer pre-scaler (1, 8, 64, 256 or 1024). The remaining data is truncated to 16-bit HW Timer.
 *                                    An example how this is converted into hardware timer is given at \ref timerUpdate function.
 *    @param[in[  limit               MOTOR_CTRL_ENDPOINT_MILLIS used in \ref control: motor rotation is stopped after limit (in ms).
 *                                    MOTOR_CTRL_ENDPOINT_MILLIS not used in control:  limits motor rotation in steps (0 = no limit at all, motor keeps on spinning).
 *    @param[in]  timerModification   Allows to variate speed by adding or subtracting this value from timerTop32virtual. (0 = none).
 *                                    The caller is responsible to ensure after \ref modCounts the \ref timerTop32Virtual result is still within the valid range.
 *    @param[in]  timerUpd_ms         Only required if MOTOR_CTRL_SPEED_VARIABLE is set in \ref control byte.
 *                                    Every x ms, the timer modification is performed (0 = none, otherwise value must be larger than 50ms).
 *                                    The timer modificatio is performed by the function \ref motorTimerUpdateCheck. This function can be called
 *                                    automatically if MOTOR_CTRL_IRQ_TIMING is set in \ref control byte. This makes the interrupt overhead for the AVR larger (worse
 *                                    behavior) but the user does not have to call the function in his/her own code. Alternative, the user may decide not to
 *                                    set MOTOR_CTRL_IRQ_TIMING in \ref control byte, but than motor speed adjustments depend on how frequent the function
 *                                    \ref motorTimerUpdateCheck is called from the user code.
 *    @param[in]  modCounts]          After this amount of modifications where introduced, the modification stops (this is timerUpd_ms * modCounts after starting).
 *                                    This is only required if MOTOR_CTRL_SPEED_VARIABLE is set in \ref control byte.
 *                                    Minimal: 50 (except when the motor speed is fixed: the value is N/A).
 *                                    Maximal: 65535ms.
 *    @retval int8_t                  -119: MOTOR_SPEED_VARIABLE set, but timerModification == 0 (if the timer is not modified, why expect variable motor speed?).
 *                                    -118: MOTOR_SPEED_VARIABLE set, but timerUpd_ms == 0 (there is no interval defined when the speed has to be adjusted).
 *                                    -117: MOTOR_SPEED_VARIABLE set, but modCounts == 0 (if there is no stop on changing motor speed, it eventually reaches the limits).
 *                                    !0: Warning:
 *                                     Bit0: timerTop32virtual != 0 and driver is disabled (if the motor is stopped, also set the virtual timer to zero).
 *                                     Bit1: timerModification != 0 and driver is disabled (if the L298 driver is not enabled, why try to change the motor speed?).
 *                                     Bit2: timerUpd_ms != 0 and driver is disabled (why change motor speed if the L298 driver is disabled?).
 *                                     Bit3: modCounts != 0 and driver is disabled (why stop motor speed changes if the L298 driver is disabled?).
 *                                     Bit6: timerUpd_ms != 0 but less than 50ms -> modified to 50ms (motor speed adjustment is limit to 20 speed changes per second).
 *                                     Bit7: 0 (if negative, it's an error code, so warnings always have this bit at zero).
 *                                    0:   No problem detected.
 */
int8_t
motorSetSpeed (uint8_t  control,
               uint32_t timerTop32virtual,                              //Virtual 32-bit timer TOP to apply motor steps; 0 = STOP motor.
               uint32_t limit,                                          //Limits amount of steps to execute (0 = never stop) or amount of millis to execute.
               uint32_t timerModification,                              //Virtual 32-bit correction on timerTop32Virtual.
               uint16_t timerUpd_ms,                                    //Every xx ms, the correction is applied.
               uint16_t modCounts)                                      //After this amount of mods has been reached, mod stops.
{
  motor_ctrl_irq = control;
  motor_timer_modify_irq = timerModification;
  motor_stat_irq = 0;
  if ((control & MOTOR_CTRL_SPEED_VARIABLE) == 0) motor_stat_irq |= MOTOR_STAT_SPIN_FIXED;
  if (timerTop32virtual == 0UL) motor_stat_irq |= MOTOR_STAT_REACHED;   //Static: position is reached.
  control = 0;                                                          //From now use control on return status byte.
  if (   (timerUpd_ms != 0)
      && (timerUpd_ms < 50))
  {
    timerUpd_ms = 50;
    control |= 0x20;                                                    //Return Warning: ms modified to 50ms.
  }
  if (   (timerTop32virtual != 0)                                       //Regardless the core clock, the minimal timer interrupt rate is 640 I/O clocks.
      && (timerTop32virtual < 640))                                     //Unless the timer is stopped (0), the minimal value is kept 640.
  {
    timerTop32virtual = 640;
    control |= 0x40;                                                    //Warning: timerTop32virtual modified because less than 640.
  }
  motor_update_time_irq = timerUpd_ms;
  motor_timer_top32virt_irq = timerTop32virtual;
  motor_modCounts_irq = modCounts;
  if ((motor_ctrl_irq & MOTOR_CTRL_ENDPOINT_MILLIS) == 0) motor_step_count_limit_irq = limit;
  else motor_millis_endtime_irq = limit ? millis () + limit : 0;
  if ((motor_ctrl_irq & MOTOR_CTRL_SPEED_VARIABLE) != 0)
  {
    if (timerModification == 0UL) return -119;                          //Virtual timer modification value == 0 -> no variable speed possible.
    if (timerUpd_ms == 0) return -118;                                  //No interval given when to update variable speed timer interval.
    if (   ((motor_ctrl_irq & MOTOR_CTRL_ENDPOINT_MILLIS) == 0)         //Endpoint in modCounts.
        && (modCounts == 0)) return -117;                               //No end condition set for variable speed when counting modifications.
  }
  if ((motor_ctrl_irq & MOTOR_CTRL_DRV_EN) == 0)
  {
    #ifdef ARDUINO_MOTOR_SHIELD
      digitalWrite (MOTOR_ENABLE_A, LOW);
      digitalWrite (MOTOR_ENABLE_B, LOW);
    #else
      MOTOR_IOPORT = ~MOTOR_OUTPUTS;                                    //Disable all drivers.
    #endif
    #ifdef MOTOR_TIMER1
      TIMSK1 &= ~(1 << TOIE1);                                          //Disable Timer Overflow Interrupt Enable.
    #endif
    #ifdef MOTOR_TIMER3
      TIMSK3 &= ~(1 << TOIE3);                                          //Disable Timer Overflow Interrupt Enable.
    #endif
    #ifdef MOTOR_TIMER4
      TIMSK4 &= ~(1 << TOIE4);                                          //Disable Timer Overflow Interrupt Enable.
    #endif
    #ifdef MOTOR_TIMER5
      TIMSK5 &= ~(1 << TOIE5);                                          //Disable Timer Overflow Interrupt Enable.
    #endif
    if (timerTop32virtual != 0) control |= 0x01;                        //Warning: virtual timer TOP32 is not zero when driver is disabled.
    if (timerModification != 0) control |= 0x02;                        //Warning: virtual timer modification value is not zero when driver is disabled.
    if (timerUpd_ms != 0) control |= 0x04;                              //Timer update rate is not zero when driver is disabled.
    if (modCounts != 0) control |= 0x08;                                //modCounts is not zero, when driver is disabled.
  }
  else
  {
    #ifdef ARDUINO_MOTOR_SHIELD
      digitalWrite (MOTOR_ENABLE_A, HIGH);
      digitalWrite (MOTOR_ENABLE_B, HIGH);                              //Enable the L298 Drivers.
    #endif
    if (timerTop32virtual != 0UL)
    {
      #ifdef MOTOR_TIMER1
        TCCR1A = (1 << WGM11) | (1 << WGM10);                           //Required setup to use Fast PWM mode.
      #endif
      #ifdef MOTOR_TIMER3
        TCCR3A = (1 << WGM31) | (1 << WGM30);                           //Required setup to use Fast PWM mode.
      #endif
      #ifdef MOTOR_TIMER4
        TCCR4A = (1 << WGM41) | (1 << WGM40);                           //Required setup to use Fast PWM mode.
      #endif
      #ifdef MOTOR_TIMER5
        TCCR5A = (1 << WGM51) | (1 << WGM50);                           //Required setup to use Fast PWM mode.
      #endif
      timerUpdate (timerTop32virtual);
      timerSkipCount (motor_timer_top32virt_irq);
      motor_time_irq = millis ();                                       //Set the start time.
      motor_stat_irq |= MOTOR_STAT_SPINNING;
      #ifdef MOTOR_TIMER1
        TIMSK1 |= (1 << TOIE1);                                         //Enable Timer Overflow Interrupt Enable.
      #endif
      #ifdef MOTOR_TIMER3
        TIMSK3 |= (1 << TOIE3);                                         //Enable Timer Overflow Interrupt Enable.
      #endif
      #ifdef MOTOR_TIMER4
        TIMSK4 |= (1 << TOIE4);                                         //Enable Timer Overflow Interrupt Enable.
      #endif
      #ifdef MOTOR_TIMER5
        TIMSK5 |= (1 << TOIE5);                                         //Enable Timer Overflow Interrupt Enable.
      #endif
    }
    else
    {
      #ifdef MOTOR_TIMER1
        TIMSK1 &= ~(1 << TOIE1);                                        //Disable Timer Overflow Interrupt Enable because 32bit vritual timer is zero.
      #endif
      #ifdef MOTOR_TIMER3
        TIMSK3 &= ~(1 << TOIE3);                                        //Disable Timer Overflow Interrupt Enable because 32bit vritual timer is zero.
      #endif
      #ifdef MOTOR_TIMER4
        TIMSK4 &= ~(1 << TOIE4);                                        //Disable Timer Overflow Interrupt Enable because 32bit vritual timer is zero.
      #endif
      #ifdef MOTOR_TIMER5
        TIMSK5 &= ~(1 << TOIE5);                                        //Disable Timer Overflow Interrupt Enable because 32bit vritual timer is zero.
      #endif
    }
  }
  return control;
}


